/* Copyright (C) 2008 and 2009 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the src/utils sub-directory);
   if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file should by reason
only of that instantiation be subject to the restrictions of use in
the GNU Lesser General Public License.  With that in mind, the words
"and instantiations of templates (of any length)" shall be treated as
inserted in the fourth paragraph of section 5 of that licence after
the words "and small inline functions (ten lines or less in length)".
This does not affect any other reason why object code may be subject
to the restrictions in that licence (nor for the avoidance of doubt
does it affect the application of section 2 of that licence to
modifications of the source code in this file).

*/

#include "prog_defs.h"      // for write_error()

#include "callback.h"
#include "emitter.h"


// uncomment the following if the specialist write_error() function in
// prog_defs.h is not available in the application in question
/*
#include <iostream>
#ifdef HAVE_OSTREAM
#include <ostream>
#endif
inline void write_error(const char* message) {
  std::cerr << message;
}
*/

namespace { // callbacks for internal use only

extern "C" {
  gboolean callback_wrapper(void*);
  gboolean tracked_callback_wrapper(void*);
}

gboolean callback_wrapper(void* data) {
  const Callback::Callback* cb = static_cast<Callback::Callback*>(data);
  try {
    cb->dispatch();
  }
  // we can't propagate exceptions from functions with C linkage
  catch (...) {
    write_error("Exception thrown in callback_wrapper() for Callback::post() function\n");
  }
  delete cb;
  return false;
}

gboolean tracked_callback_wrapper(void* data) {
  const Callback::SafeEmitter* e = static_cast<Callback::SafeEmitter*>(data);
  try {
    e->emit();
  }
  // we can't propagate exceptions from functions with C linkage
  catch (...) {
    write_error("Exception thrown in tracked_callback_wrapper() for Callback::post() function\n");
  }
  delete e;
  return false;
}

} // anonymous namespace

namespace Callback {

void post(const Callback* cb, gint priority) {
  g_idle_add_full(priority, callback_wrapper, const_cast<Callback*>(cb), 0);
}

void post(const Callback* cb, Releaser& r, gint priority) {
  SafeEmitter* e = new SafeEmitter;
  e->connect(cb, r);
  g_idle_add_full(priority, tracked_callback_wrapper, e, 0);
}

} // namespace Callback
